#lang scheme

(require "state-monad.ss")

;; -- TYPES -- ;;

; node type
; id : number
; outgoing-nodes : listof-number
(define-struct node (id outgoing-nodes))

; edge : (node-id . node-id)

; dfs-state type
; pre-post : hash-from-node-id-to-number-pair
; tree-edges : set-of-edge
; visited : set-of-node-id
; next-num : number
(define-struct dfs-state (pre-post tree-edges visited next-num))

; starting-dfs-state : graph -> dfs-state
(define (starting-dfs-state graph)
  (make-dfs-state (make-immutable-hash
                   (map (lambda (x) (cons x (cons #f #f))) (hash-keys graph)))
                  (make-immutable-hash empty)
                  (make-immutable-hash empty) 
                  1))

; graph : hash-from-node-id-to-node

;; -- Set type and helpers -- ;;

; set a : hash-from-a-to-boolean

; list-to-set : listof-a -> setof-a
(define (list-to-set lst)
  (make-immutable-hash (map (lambda (x) (cons x #t)) lst)))

; set-member : a setof-a -> boolean
(define (set-member e set)
  (hash-has-key? set e))

; set-insert : a setof-a -> setof-a
(define (set-insert e set)
  (hash-set set e #t))
           
; hash-keys : hash-from-a-to-? -> listof-a
(define (hash-keys hash) (for/list ([x (in-hash-keys hash)]) x))

;; -- MONAD HELPERS -- ;;

;; -- pre-post -- ;;

(define get-pre-post
  (>>= get (lambda (s) (return (dfs-state-pre-post s)))))
(define (set-pre-post pre-post)
  (>>= get (lambda (s)
  (put (struct-copy dfs-state s [pre-post pre-post])))))
(define (mark-pre-number node-id)
  (do
    (<- pre-post get-pre-post)
    (<- next-num get-next-num-incr)
    (set-pre-post (hash-update pre-post node-id (lambda (x) (cons next-num (cdr x)))))))
(define (mark-post-number node-id)
  (do
    (<- pre-post get-pre-post)
    (<- next-num get-next-num-incr)
    (set-pre-post (hash-update pre-post node-id (lambda (x) (cons (car x) next-num))))))

;; -- visited -- ;;

(define get-visited
  (>>= get (lambda (s) (return (dfs-state-visited s)))))
(define (set-visited visited)
  (>>= get (lambda (s)
  (put (struct-copy dfs-state s [visited visited])))))
(define (already-visited-node node-id)
  (do
    (<- visited get-visited)
    (return (set-member node-id visited))))
(define (mark-visited node-id)
  (do
    (<- visited get-visited)
    (set-visited (set-insert node-id visited))))

;; -- tree-edges -- ;;

(define get-tree-edges
  (>>= get (lambda (s) (return (dfs-state-tree-edges s)))))
(define (set-tree-edges tree-edges)
  (>>= get (lambda (s)
  (put (struct-copy dfs-state s [tree-edges tree-edges])))))
(define (add-tree-edge edge)
  (do
    (<- tree-edges get-tree-edges)
    (set-tree-edges (set-insert edge tree-edges))))

;; -- next-num -- ;;

(define get-next-num-incr
  (>>= get (lambda (s)
  (>>= (put (struct-copy dfs-state 
                         s [next-num (+ 1 (dfs-state-next-num s))]))
       (lambda (_) 
  (return (dfs-state-next-num s)))))))

;; -- DFS -- ;;

; dfs : graph -> dfs-state
; dfs pseudo-code:
; dfs graph =
;   for-each node in graph
;     visit node
(define (dfs graph)
  (exec-state (for-each-M (hash-keys graph) 
                          (lambda (id) (visit id #f graph))) 
              (starting-dfs-state graph)))

; visit : node-id node-id-or-false graph -> State dfs-state ()
; visit pseudo-code:
; visit node-id came-from graph =
;  if not already-visited node then
;    mark-visited node-id
;    if came-from then add-tree-edge (came-from . node-id)
;    mark-pre-number node-id
;    for child-id in node-children node:
;      visit child-id node-id graph
;    mark-post-number node-id
(define (visit node-id came-from graph)
  (do
    (<- already-visited (already-visited-node node-id))
    (unless already-visited
      (do 
        (mark-visited node-id)
        (when came-from (add-tree-edge (cons came-from node-id)))
        (mark-pre-number node-id)
        (for-each-M (node-outgoing-nodes (hash-ref graph node-id)) 
                    (lambda (child-id) (visit child-id node-id graph)))
        (mark-post-number node-id)))))
               
;; -- PARSING HELPERS -- ;;

(define (make-graph sexp)
  (cond
    [(empty? sexp) (make-immutable-hash empty)]
    [else (let* ((unparsed-node (first sexp))
                 (node-id (first unparsed-node))
                 (node-outgoing (rest unparsed-node))
                 (node (make-node node-id node-outgoing)))
            (hash-set (make-graph (rest sexp)) node-id node))]))

;; -- TESTING HELPERS -- ;;

(define (test a b)
  (if (equal? a b) 
      (void)
      (error "test failed")))

;; -- TESTS -- ;;

(define graph1 (make-graph '((1 2 3) (2 3) (3))))
(test (dfs-state-pre-post (dfs graph1)) 
      #hash((1 . (1 . 6))
            (2 . (2 . 5)) 
            (3 . (3 . 4))))
(test (dfs-state-tree-edges (dfs graph1))
      #hash(((1 . 2) . #t)
            ((2 . 3) . #t)))

(define graph2 (make-graph '((1 2 3) (2) (3) (4 5 6 7) (5 7) (6 7) (7))))
(test (dfs-state-pre-post (dfs graph2)) 
      #hash((1 . (1 . 6)) 
            (2 . (2 . 3))
            (3 . (4 . 5)) 
            (4 . (7 . 14)) 
            (5 . (8 . 11))
            (6 . (12 . 13))
            (7 . (9 . 10))))
(test (dfs-state-tree-edges (dfs graph2))
      #hash(((1 . 2) . #t)
            ((1 . 3) . #t)
            ((4 . 5) . #t)
            ((4 . 6) . #t)
            ((5 . 7) . #t)))